from sklearn.datasets import make_blobs
from sklearn.model_selection import train_test_split
from reval.best_nclust_cv import FindBestClustCV
from reval.visualization import plot_metrics
from sklearn.neighbors import KNeighborsClassifier
from sklearn.cluster import AgglomerativeClustering
from sklearn.metrics import adjusted_mutual_info_score, zero_one_loss
from reval.relative_validation import _kuhn_munkres_algorithm
import matplotlib.pyplot as plt
import numpy as np

# NUMBER OF FEATURES
# ------------------
# The first example illustrates how results can be influenced by the number of dataset features.
# Increasing the number of features can fix the problem.

data1 = make_blobs(1000, 10, 5, cluster_std=5, random_state=42)

# Plot synthetic dataset
plt.scatter(data1[0][:, 0], data1[0][:, 1],
            c=data1[1], cmap='rainbow_r')
plt.title('True labels for 10-feature dataset')

X_tr, X_ts, y_tr, y_ts = train_test_split(data1[0],
                                          data1[1],
                                          test_size=0.30, random_state=42,
                                          stratify=data1[1])
# Apply relative clustering validation with KNN and Hierarchical clustering
classifier = KNeighborsClassifier()
clustering = AgglomerativeClustering()

findbestclust = FindBestClustCV(nfold=10, nclust_range=[2, 7],
                                s=classifier, c=clustering, nrand=100)
metrics, nbest, _ = findbestclust.best_nclust(X_tr, y_tr)
out = findbestclust.evaluate(X_tr, X_ts, nbest)

plot_metrics(metrics, "Reval performance for synthetic dataset with 10 features")

plt.scatter(X_ts[:, 0], X_ts[:, 1],
            c=out.test_cllab, cmap='rainbow_r')
plt.title("Predicted labels for 10-feature dataset")

# Compare Reval solution to true labels
print(f'AMI test set = {adjusted_mutual_info_score(y_ts, out.test_cllab)}')
relabeling = _kuhn_munkres_algorithm(y_ts, out.test_cllab)
print(f'ACC test set = {1 - zero_one_loss(y_ts, relabeling)}')

# Increase the number of features from 10 to 20
data2 = make_blobs(1000, 20, 5, cluster_std=5, random_state=42)

# Plot synthetic dataset
plt.scatter(data2[0][:, 0],
            data2[0][:, 1],
            c=data2[1],
            cmap='rainbow_r')
plt.title('True labels for 20-feature dataset')

X_tr, X_ts, y_tr, y_ts = train_test_split(data2[0],
                                          data2[1],
                                          test_size=0.30, random_state=42,
                                          stratify=data2[1])

findbestclust = FindBestClustCV(nfold=10, nclust_range=[2, 7],
                                s=classifier, c=clustering, nrand=100)
metrics, nbest, _ = findbestclust.best_nclust(X_tr, y_tr)
out = findbestclust.evaluate(X_tr, X_ts, nbest)

plot_metrics(metrics, "Reval performance for synthetic dataset with 20 features")

plt.scatter(X_ts[:, 0], X_ts[:, 1],
            c=out.test_cllab, cmap='rainbow_r')
plt.title("Predicted labels for 20-feature dataset")

# Compare predicted labels with true labels (Adjusted Mutual Information, accuracy)
print(f'AMI test set = {adjusted_mutual_info_score(y_ts, out.test_cllab)}')
relabeling = _kuhn_munkres_algorithm(y_ts, out.test_cllab)
print(f'ACC test set = {1 - zero_one_loss(y_ts, relabeling)}')

# NUMBER OF SAMPLES
# -----------------
np.random.seed(42)

# We generate three random samples from normal distributions
data1 = np.random.normal(-5, size=(100, 2))
data2 = np.random.normal(12, 2.5, size=(50, 2))
data3 = np.random.normal(6, 2.5, size=(50, 2))
data = np.append(data1, data2, axis=0)
data = np.append(data, data3, axis=0)

label = [0] * 100 + [1] * 50 + [2] * 50

plt.scatter(data[:, 0], data[:, 1],
            c=label, cmap='rainbow_r')
plt.title('Random samples from normal distribution Ns=(100, 50, 50)')

classifier = KNeighborsClassifier()
clustering = AgglomerativeClustering()

X_tr, X_ts, y_tr, y_ts = train_test_split(data, label,
                                          test_size=0.30,
                                          random_state=42,
                                          stratify=label)

# Apply relative clustering validation with KNN and Hierarchical clustering
findbestclust = FindBestClustCV(nfold=10, nclust_range=[2, 7],
                                s=classifier, c=clustering, nrand=100)
metrics, nbest, _ = findbestclust.best_nclust(X_tr, y_tr)
out = findbestclust.evaluate(X_tr, X_ts, nbest)
plot_metrics(metrics, "Reval performance for synthetic dataset with Ns=(100, 50, 50)")

plt.scatter(X_ts[:, 0], X_ts[:, 1],
            c=_kuhn_munkres_algorithm(np.array(y_ts),
                                      out.test_cllab),
            cmap='rainbow_r')
plt.title(f'Predicted labels for classes with Ns=(100, 50, 50)')

# We now increase the number of samples in groups 2 and 3 to 500
data1 = np.random.normal(-5, size=(100, 2))
data2 = np.random.normal(12, 2.5, size=(500, 2))
data3 = np.random.normal(6, 2.5, size=(500, 2))
data = np.append(data1, data2, axis=0)
data = np.append(data, data3, axis=0)

label = [0] * 100 + [1] * 500 + [2] * 500

plt.scatter(data[:, 0], data[:, 1],
            c=label, cmap='rainbow_r')
plt.title('Random samples from normal distribution Ns=(100, 500, 500)')

classifier = KNeighborsClassifier()
clustering = AgglomerativeClustering()

X_tr, X_ts, y_tr, y_ts = train_test_split(data, label,
                                          test_size=0.30,
                                          random_state=42,
                                          stratify=label)

# Apply relative clustering validation with KNN and Hierarchical clustering
findbestclust = FindBestClustCV(nfold=10, nclust_range=[2, 7],
                                s=classifier, c=clustering, nrand=100)
metrics, nbest, _ = findbestclust.best_nclust(X_tr, y_tr)
out = findbestclust.evaluate(X_tr, X_ts, nbest)
plot_metrics(metrics, "Reval performance for synthetic dataset with Ns=(100, 500, 500)")

plt.scatter(X_ts[:, 0], X_ts[:, 1],
            c=y_ts,
            cmap='rainbow_r')
plt.title(f'Test set true labels for classes with Ns=(100, 500, 500)')

plt.scatter(X_ts[:, 0], X_ts[:, 1],
            c=_kuhn_munkres_algorithm(np.array(y_ts),
                                      out.test_cllab),
            cmap='rainbow_r')
plt.title(f'Predicted labels for classes with Ns=(100, 500, 500)')

# Performance scores
# Test set ACC
print(f'Test set external '
      f'ACC = {1 - zero_one_loss(y_ts, _kuhn_munkres_algorithm(np.array(y_ts), out.test_cllab))}')
print(f'AMI = {adjusted_mutual_info_score(y_ts, out.test_cllab)}')
print(f"Validation stability metrics: {metrics['val'][nbest]}")
print(f"Test set ACC = {out.test_acc}")



from sklearn.datasets import make_blobs
from sklearn.model_selection import train_test_split
from reval.best_nclust_cv import FindBestClustCV
from sklearn.neighbors import KNeighborsClassifier
from sklearn.cluster import AgglomerativeClustering
from sklearn.metrics import zero_one_loss, adjusted_mutual_info_score
from reval.visualization import plot_metrics
import matplotlib.pyplot as plt
from reval.utils import kuhn_munkres_algorithm

# Generate 1,000 samples for 5 blobs
# ----------------------------------
data = make_blobs(1000, 2, 5, random_state=42)
plt.scatter(data[0][:, 0],
            data[0][:, 1],
            c=data[1], cmap='rainbow_r')

# Split them into training and test set (30% of data)
X_tr, X_ts, y_tr, y_ts = train_test_split(data[0],
                                          data[1],
                                          test_size=0.30,
                                          random_state=42,
                                          stratify=data[1])

# Apply relative clustering validation with KNN and Hierarchical clustering
classifier = KNeighborsClassifier()
clustering = AgglomerativeClustering()

findbestclust = FindBestClustCV(nfold=10,
                                nclust_range=[2, 7],
                                s=classifier,
                                c=clustering,
                                nrand=100)
metrics, nbest, _ = findbestclust.best_nclust(X_tr, y_tr)
out = findbestclust.evaluate(X_tr, X_ts, nbest)

perm_lab = kuhn_munkres_algorithm(y_ts, out.test_cllab)

print(f"Best number of clusters: {nbest}")
print(f"Test set external ACC: "
      f"{1 - zero_one_loss(y_ts, perm_lab)}")
print(f'AMI = {adjusted_mutual_info_score(y_ts, out.test_cllab)}')
print(f"Validation set normalized stability (misclassification): {metrics['val'][nbest]}")
print(f'Test set ACC = {out.test_acc}')

plot_metrics(metrics, title="Reval performance")

plt.scatter(X_ts[:, 0], X_ts[:, 1],
            c=y_ts, cmap='rainbow_r')
plt.title("True labels for test set")

plt.scatter(X_ts[:, 0], X_ts[:, 1],
            c=perm_lab, cmap='rainbow_r')
plt.title("Clustering labels for test set")

# Create a noisy dataset with 5 clusters
# ----------------------------------------
data_noisy = make_blobs(1000, 10, 5, random_state=42, cluster_std=3)
plt.scatter(data_noisy[0][:, 0],
            data_noisy[0][:, 1],
            c=data_noisy[1],
            cmap='rainbow_r')

Xnoise_tr, Xnoise_ts, ynoise_tr, ynoise_ts = train_test_split(data_noisy[0],
                                                              data_noisy[1],
                                                              test_size=0.30,
                                                              random_state=42,
                                                              stratify=data_noisy[1])

metrics_noise, nbest_noise, _ = findbestclust.best_nclust(Xnoise_tr, ynoise_tr)
out_noise = findbestclust.evaluate(Xnoise_tr, Xnoise_ts, nbest_noise)

plot_metrics(metrics_noise, title="Reval performance")

perm_lab_noise = kuhn_munkres_algorithm(ynoise_ts, out_noise.test_cllab)

print(f"Best number of clusters: {nbest_noise}")
print(f"Test set external ACC: "
      f"{1 - zero_one_loss(ynoise_ts, perm_lab_noise)}")
print(f'AMI = {adjusted_mutual_info_score(ynoise_ts, out_noise.test_cllab)}')
print(f"Validation set normalized stability (misclassification): {metrics_noise['val'][nbest_noise]}")
print(f"Result accuracy (on test set): "
      f"{out_noise.test_acc}")

plt.scatter(Xnoise_ts[:, 0], Xnoise_ts[:, 1],
            c=ynoise_ts, cmap='rainbow_r')
plt.title("True labels")

plt.scatter(Xnoise_ts[:, 0], Xnoise_ts[:, 1],
            c=perm_lab_noise, cmap='rainbow_r')
plt.title("Clustering labels for test set")

# Pre-processing with UMAP
from umap import UMAP

transform = UMAP(n_components=10, n_neighbors=30, min_dist=0.0)

Xtr_umap = transform.fit_transform(Xnoise_tr)
Xts_umap = transform.transform(Xnoise_ts)

plt.scatter(Xtr_umap[:, 0], Xtr_umap[:, 1],
            c=ynoise_tr, cmap='rainbow_r')
plt.title("UMAP-transformed training set with true labels")

plt.scatter(Xts_umap[:, 0], Xts_umap[:, 1],
            c=ynoise_ts, cmap='rainbow_r')
plt.title("UMAP-transformed test set with true labels")

metrics, nbest, _ = findbestclust.best_nclust(Xtr_umap, ynoise_tr)
out = findbestclust.evaluate(Xtr_umap, Xts_umap, nbest)

plot_metrics(metrics, title='Reval performance of UMAP-transformed dataset')

perm_noise = kuhn_munkres_algorithm(ynoise_ts, out.test_cllab)

print(f"Best number of clusters: {nbest}")
print(f"Test set external ACC: "
      f"{1 - zero_one_loss(ynoise_ts, perm_noise)}")
print(f'AMI = {adjusted_mutual_info_score(ynoise_ts, out.test_cllab)}')
print(f"Validation set normalized stability (misclassification): {metrics['val'][nbest]}")
print(f"Result accuracy (on test set): "
      f"{out.test_acc}")

plt.scatter(Xts_umap[:, 0], Xts_umap[:, 1],
            c=perm_noise, cmap='rainbow_r')
plt.title("Predicted labels for UMAP-preprocessed test set")



import unittest
import numpy as np
from reval import relative_validation
from reval.relative_validation import RelativeValidation

from sklearn.neighbors import KNeighborsClassifier
from sklearn.cluster import AgglomerativeClustering

# Modify to test other functions and parameters
RNDLABELS_ITER = 100
CLASSIFIER = KNeighborsClassifier(n_neighbors=15)
CLUSTERING = AgglomerativeClustering(n_clusters=2)
NCLUST_RANGE = [2, 20]
NFOLD = 10


class TestReval(unittest.TestCase):
    @classmethod
    def setUp(cls):
        cls.s = CLASSIFIER
        cls.c = CLUSTERING
        cls.nrand = RNDLABELS_ITER
        cls.reval_cls = RelativeValidation(s=cls.s, c=cls.c, nrand=cls.nrand)

    def test_revaltraining(self):
        data_tr = np.array([[0] * 10,
                            [1] * 10] * 20)
        misclass, model, labels = self.reval_cls.train(data_tr)
        self.assertSequenceEqual([misclass] + [labels.tolist()],
                                 [0.0, relative_validation._kuhn_munkres_algorithm(labels,
                                                                                   np.array([1, 0] * 20)).tolist()])
        self.assertEqual(type(model), type(self.s))

    def test_revaltest(self):
        data_tr = np.array([[0] * 10,
                            [1] * 10] * 20)
        data_ts = np.array([[0] * 10,
                            [1] * 10] * 10)
        model = self.s.fit(data_tr, [1, 0] * 20)
        misclass, labels = self.reval_cls.test(data_ts, model)
        self.assertSequenceEqual([misclass] + [labels.tolist()],
                                 [0.0, [1, 0] * 10])

    def test_rndlabels(self):
        data_tr = np.array([[0] * 10,
                            [1] * 10] * 20)
        labels_tr = np.array([0, 1] * 20)
        data_ts = np.array([[0] * 10,
                            [1] * 10] * 10)
        labels_ts = np.array([0, 1] * 10)
        misclass_ts = self.reval_cls.rndlabels_traineval(data_tr,
                                                         data_ts,
                                                         labels_tr,
                                                         labels_ts)
        self.assertTrue(misclass_ts > 0)

    def test_khun_munkres_algorithm(self):
        true_lab = np.array([1, 1, 1, 0, 0, 0])
        pred_lab = np.array([0, 0, 0, 1, 1, 1])
        new_lab = relative_validation._kuhn_munkres_algorithm(true_lab, pred_lab)
        self.assertSequenceEqual(new_lab.tolist(), [1, 1, 1, 0, 0, 0])


if __name__ == 'main':
    unittest.main()



from sklearn.datasets import fetch_openml
from sklearn.model_selection import train_test_split
from reval.best_nclust_cv import FindBestClustCV
from sklearn.neighbors import KNeighborsClassifier
from sklearn.cluster import AgglomerativeClustering
from sklearn.metrics import zero_one_loss, adjusted_mutual_info_score
import matplotlib.pyplot as plt
from umap import UMAP
from reval.visualization import plot_metrics
from reval.relative_validation import _kuhn_munkres_algorithm

# MNIST dataset with 10 classes
mnist, label = fetch_openml('mnist_784', version=1, return_X_y=True)
transform = UMAP(n_neighbors=30, min_dist=0.0, n_components=10, random_state=42)

# Stratified subsets of 7000 elements for both training and test set
mnist_tr, mnist_ts, label_tr, label_ts = train_test_split(mnist, label,
                                                          train_size=0.1,
                                                          test_size=0.1,
                                                          random_state=42,
                                                          stratify=label)

# Dimensionality reduction with UMAP as pre-processing step
mnist_tr = transform.fit_transform(mnist_tr)
mnist_ts = transform.transform(mnist_ts)

plt.scatter(mnist_tr[:, 0],
            mnist_tr[:, 1],
            c=label_tr.astype(int),
            s=0.1,
            cmap='rainbow_r')
plt.title('UMAP-transformed training subsample of MNIST dataset (N=7,000)')

plt.scatter(mnist_ts[:, 0], mnist_ts[:, 1],
            c=label_ts.astype(int), s=0.1, cmap='rainbow_r')
plt.title('UMAP-transformed test subsample of MNIST dataset (N=7,000)')

# Run relative clustering validation
classifier = KNeighborsClassifier()
clustering = AgglomerativeClustering()

findbestclust = FindBestClustCV(nfold=10, nclust_range=[2, 12],
                                s=classifier, c=clustering, nrand=100)

metrics, nbest, _ = findbestclust.best_nclust(mnist_tr, label_tr)
out = findbestclust.evaluate(mnist_tr, mnist_ts, nbest)

plot_metrics(metrics, "Relative clustering validation performance on MNIST dataset")

perm_lab = _kuhn_munkres_algorithm(label_ts.astype(int), out.test_cllab)

plt.scatter(mnist_ts[:, 0], mnist_ts[:, 1],
            c=perm_lab, s=0.1, cmap='rainbow_r')
plt.title("Predicted labels for MNIST test set")

print(f"Best number of clusters: {nbest}")
print(f"Test set external ACC: "
      f"{1 - zero_one_loss(label_ts.astype(int), perm_lab)}")
print(f'AMI = {adjusted_mutual_info_score(label_ts.astype(int), perm_lab)}')
print(f"Validation set normalized stability (misclassification): {metrics['val'][nbest]}")
print(f"Result accuracy (on test set): "
      f"{out.test_acc}")



from sklearn.datasets import make_blobs, load_digits
from sklearn.model_selection import train_test_split
from reval.best_nclust_cv import FindBestClustCV
from sklearn.neighbors import KNeighborsClassifier
from sklearn.cluster import AgglomerativeClustering, KMeans, SpectralClustering
from sklearn.metrics import zero_one_loss, adjusted_mutual_info_score
from reval.visualization import plot_metrics
import matplotlib.pyplot as plt
from reval.utils import kuhn_munkres_algorithm
from umap import UMAP
from sklearn.ensemble import RandomForestClassifier
from sklearn.svm import SVC
from sklearn.linear_model import LogisticRegression
import numpy as np
from datasets.manuscript_builddatasets import build_ucidatasets
import warnings
import logging

warnings.filterwarnings('ignore')
logging.basicConfig(filename='manuscript_examples.log',
                    filemode='a',
                    format='%(asctime)s, %(levelname)s %(message)s',
                    datefmt='%H:%M:%S',
                    level=logging.INFO)

"""
Three example functions that can be run from shell (change main at the bottom). 

Example 1: blobs dataset;
Example 2: real-world dataset (handwritten digits)
Example 3: ensamble learning with 18 datasets from UCI Machine Learning Repository.
           Algorithm performances are saved to manuscript_example.log file.
"""


# EXAMPLE 1: Isotropic Gaussian blobs
def example_1():
    data = make_blobs(1000, 2, 5, center_box=(-20, 20),
                      random_state=42)
    plt.figure(figsize=(6, 4))
    plt.scatter(data[0][:, 0],
                data[0][:, 1],
                c=data[1], cmap='rainbow_r')
    plt.title("Blobs dataset (N=1000)")
    plt.show()

    X_tr, X_ts, y_tr, y_ts = train_test_split(data[0],
                                              data[1],
                                              test_size=0.30,
                                              random_state=42,
                                              stratify=data[1])

    classifier = KNeighborsClassifier(n_neighbors=5)
    clustering = KMeans()

    findbestclust = FindBestClustCV(nfold=10,
                                    nclust_range=[2, 7],
                                    s=classifier,
                                    c=clustering,
                                    nrand=100)
    metrics, nbest, _ = findbestclust.best_nclust(X_tr, iter_cv=10, strat_vect=y_tr)
    out = findbestclust.evaluate(X_tr, X_ts, nbest)

    perm_lab = kuhn_munkres_algorithm(y_ts, out.test_cllab)

    print(f"Best number of clusters: {nbest}")
    print(f"Test set prediction ACC: "
          f"{1 - zero_one_loss(y_ts, perm_lab)}")
    print(f'AMI (true labels vs predicted labels) = '
          f'{adjusted_mutual_info_score(y_ts, out.test_cllab)}')
    print(f"Validation set normalized stability (misclassification):"
          f"{metrics['val'][nbest]}")
    print(f'Test set ACC = {out.test_acc} '
          f'(true labels vs predicted labels)')

    plot_metrics(metrics, title="Reval performance blobs dataset",
                 legend_loc=2)

    plt.figure(figsize=(6, 4))
    plt.scatter(X_ts[:, 0], X_ts[:, 1],
                c=y_ts, cmap='rainbow_r')
    plt.title("Test set true labels (blobs dataset)")
    plt.show()

    plt.figure(figsize=(6, 4))
    plt.scatter(X_ts[:, 0], X_ts[:, 1],
                c=perm_lab, cmap='rainbow_r')
    plt.title("Test set clustering labels (blobs dataset)")
    plt.show()


# EXAMPLE 2: Handwritten digits dataset example
def example_2():
    digits_dataset = load_digits()

    digits_data = digits_dataset['data']
    digits_target = digits_dataset['target']

    X_tr, X_ts, y_tr, y_ts = train_test_split(digits_data,
                                              digits_target,
                                              test_size=0.40,
                                              random_state=42,
                                              stratify=digits_target)

    transform = UMAP(n_components=2,
                     random_state=42,
                     n_neighbors=30,
                     min_dist=0.0)
    X_tr = transform.fit_transform(X_tr)
    X_ts = transform.transform(X_ts)

    s = KNeighborsClassifier(n_neighbors=30)
    c = KMeans()

    reval = FindBestClustCV(s=s,
                            c=c,
                            nfold=5,
                            nclust_range=[2, 15],
                            nrand=100)

    metrics, nclustbest, _ = reval.best_nclust(X_tr, iter_cv=10, strat_vect=y_tr)

    plot_metrics(metrics, title='Reval performance digits dataset')

    out = reval.evaluate(X_tr, X_ts, nclust=nclustbest)
    perm_lab = kuhn_munkres_algorithm(y_ts, out.test_cllab)

    print(f"Best number of clusters: {nclustbest}")
    print(f"Test set prediction ACC: "
          f"{1 - zero_one_loss(y_ts, perm_lab)}")
    print(f'AMI (true labels vs predicted labels) = '
          f'{adjusted_mutual_info_score(y_ts, out.test_cllab)}')
    print(f"Validation set normalized stability (misclassification):"
          f"{metrics['val'][nclustbest]}")
    print(f'Test set ACC = {out.test_acc} '
          f'(true labels vs predicted labels)')

    plt.figure(figsize=(6, 4))
    plt.scatter(X_ts[:, 0],
                X_ts[:, 1],
                c=y_ts, cmap='rainbow_r')
    plt.title("Test set true labels (digits dataset)")
    plt.show()

    plt.figure(figsize=(6, 4))
    plt.scatter(X_ts[:, 0],
                X_ts[:, 1],
                c=perm_lab, cmap='rainbow_r')
    plt.title("Test set clustering labels (digits dataset)")
    plt.show()


# Example 3: ensemble learning
def example_3():
    # Classifiers
    knn = KNeighborsClassifier(n_neighbors=1,
                               metric='euclidean')
    rf = RandomForestClassifier(n_estimators=100,
                                random_state=42)
    svm = SVC(C=1,
              random_state=42)
    logreg = LogisticRegression(solver='liblinear',
                                random_state=42)

    classifiers = [knn, logreg, svm, rf]

    # Clustering
    hc = AgglomerativeClustering()
    km = KMeans(random_state=42)
    sc = SpectralClustering(random_state=42)

    clustering = [hc, km, sc]

    # scaler = StandardScaler()
    transform = UMAP(n_neighbors=30, min_dist=0.0, random_state=42)

    # Import benchmark datasets
    uci_data = build_ucidatasets()
    # Run ensemble learning algorithm
    for data, name in zip(uci_data, uci_data._fields):
        nclass = len(np.unique(data['target']))
        logging.info(f"Processing dataset {name}")
        logging.info(f"Number of classes: {nclass}\n")
        X_tr, X_ts, y_tr, y_ts = train_test_split(data['data'],
                                                  data['target'],
                                                  test_size=0.40,
                                                  random_state=42,
                                                  stratify=data['target'])
        X_tr = transform.fit_transform(X_tr)
        X_ts = transform.transform(X_ts)
        for s in classifiers:
            if type(s) == type(svm):
                svm.gamma = 1 / data['data'].shape[0]
            for c in clustering:
                logging.info(f"Clustering algorithm: {c} -- Classification algorithm {s}")
                reval = FindBestClustCV(s=s,
                                        c=c,
                                        nfold=5,
                                        nclust_range=[2, nclass + 3],
                                        nrand=100)
                metrics, nclustbest, _ = reval.best_nclust(X_tr, strat_vect=y_tr)

                out = reval.evaluate(X_tr, X_ts, nclust=nclustbest)
                perm_lab = kuhn_munkres_algorithm(y_ts, out.test_cllab)

                logging.info(f"Best number of clusters: {nclustbest}")
                logging.info(f"Test set prediction ACC: "
                             f"{1 - zero_one_loss(y_ts, perm_lab)}")
                logging.info(f'AMI (true labels vs predicted labels) = '
                             f'{adjusted_mutual_info_score(y_ts, out.test_cllab)}')
                logging.info(f"Validation set normalized stability (misclassification):"
                             f"{metrics['val'][nclustbest]}")
                logging.info(f'Test set ACC = {out.test_acc} '
                             f'(true labels vs predicted labels)\n')
        logging.info('*' * 100)
        logging.info('\n\n')


if __name__ == "__main__":
    example_1()
    example_2()



import numpy as np
from sklearn.metrics import zero_one_loss
from reval.utils import kuhn_munkres_algorithm


class RelativeValidation:
    """
    This class allows to perform the relative clustering validation procedure.
    A supervised algorithm is required to test cluster stability.
    Labels output from a clustering algorithm are used as true labels.

    :param s: initialized class for the supervised method
    :type s: class
    :param c: initialized class for clustering algorithm
    :type c: class
    :param nrand: number of iterations to normalize cluster stability
    :type nrand: int
    """

    def __init__(self, s, c, nrand):
        """
        Construct method
        """
        self.class_method = s
        self.clust_method = c
        self.nrand = nrand

    def train(self, train_data):
        """
        Method that performs training. It compares the clustering labels on training set
        (i.e., A(X) computed by :class:`reval.relative_validation.RelativeValidation.clust_method`) against
        the labels obtained from the classification algorithm
        (i.e., f(X), computed by :class:`reval.relative_validation.RelativeValidation.class_method`).
        It returns the misclassification error, the supervised model fitted to the data,
        and both clustering and classification labels.

        :param train_data: training dataset
        :type train_data: ndarray, (n_samples, n_features)
        :return: misclassification error, fitted supervised model object, clustering and classification labels
        :rtype: float, object, dictionary of ndarray (n_samples,)
        """

        clustlab_tr = self.clust_method.fit_predict(train_data)  # A_k(X)
        fitclass_tr = self.class_method.fit(train_data, clustlab_tr)
        classlab_tr = fitclass_tr.predict(train_data)
        misclass = zero_one_loss(clustlab_tr, classlab_tr)
        return misclass, fitclass_tr, clustlab_tr

    def test(self, test_data, fit_model):
        """
        Method that compares test set clustering labels (i.e., A(X'), computed by
        :class:`reval.relative_validation.RelativeValidation.clust_method`) against
        the (permuted) labels obtained through the classification algorithm fitted to the training set
        (i.e., f(X'), computed by
        :class:`reval.relative_validation.RelativeValidation.class_method`).
        It returns the misclassification error, together with
        both clustering and classification labels.

        :param test_data: test dataset
        :type test_data: ndarray, (n_samples, n_features)
        :param fit_model: fitted supervised model
        :type fit_model: class
        :return: misclassification error, clustering and classification labels
        :rtype: float, dictionary of ndarrays (n_samples,)
        """
        clustlab_ts = self.clust_method.fit_predict(test_data)  # A_k(X')
        classlab_ts = fit_model.predict(test_data)
        bestperm = kuhn_munkres_algorithm(classlab_ts, clustlab_ts)  # array of integers
        misclass = zero_one_loss(classlab_ts, bestperm)
        return misclass, bestperm

    def rndlabels_traineval(self, train_data, test_data, train_labels, test_labels):
        """
        Method that performs random labeling on the training set
        (N times according to
        :class:`reval.relative_validation.RelativeValidation.nrand` instance attribute) and evaluates
        the fitted models on test set.

        :param train_data: training dataset
        :type train_data: ndarray, (n_samples, n_features)
        :param test_data: test dataset
        :type test_data: ndarray, (n_samples, n_features)
        :param train_labels: training set clustering labels
        :type train_labels: ndarray, (n_samples,)
        :param test_labels: test set clustering labels
        :type test_labels: ndarray, (n_samples,)
        :return: averaged misclassification error on the test set
        :rtype: float
        """
        np.random.seed(0)
        shuf_tr = [np.random.permutation(train_labels)
                   for _ in range(self.nrand)]
        misclass_ts = list(map(lambda x: self._rescale_score_(train_data, test_data, x, test_labels), shuf_tr))
        return np.mean(misclass_ts)

    def _rescale_score_(self, xtr, xts, randlabtr, labts):
        """
        Private method that computes the misclassification error when predicting test labels
        with classification model fitted on training set with random labels.

        :param xtr: training dataset
        :type xtr: ndarray, (n_samples, n_features)
        :param xts: test dataset
        :type xts: ndarray, (n_samples, n_features)
        :param randlabtr: random labels
        :type randlabtr: ndarray, (n_samples,)
        :param labts: test set labels
        :type labts: ndarray, (n_samples,)
        :return: misclassification error
        :rtype: float
        """
        self.class_method.fit(xtr, randlabtr)
        pred_lab = self.class_method.predict(xts)
        me_ts = zero_one_loss(pred_lab, kuhn_munkres_algorithm(pred_lab, labts))
        return me_ts



import time
from sklearn.datasets import make_blobs
from sklearn.model_selection import train_test_split
from reval.best_nclust_cv import FindBestClustCV
from sklearn.neighbors import KNeighborsClassifier
from sklearn.cluster import KMeans
import matplotlib.pyplot as plt
import numpy as np


def blobs_performance():
    """
    Function performing multiple iterations of reval on simulated 5-blob datasets
    with varying number of samples and features and 10 repetitions of 10-fold CVs.
    The function plots the performance (in seconds) of the algorithm for each
    parameter configuration.
    """
    feat = [10, 100, 500, 1000]
    samples = [100, 500, 1000, 1500, 2000]

    perftime = []
    for s in samples:
        perf = []
        for f in feat:
            start = time.time()
            data = make_blobs(s, f, 5, center_box=(-20, 20),
                              random_state=42)

            X_tr, X_ts, y_tr, y_ts = train_test_split(data[0],
                                                      data[1],
                                                      test_size=0.30,
                                                      random_state=42,
                                                      stratify=data[1])

            classifier = KNeighborsClassifier(n_neighbors=5)
            clustering = KMeans()

            findbestclust = FindBestClustCV(nfold=10,
                                            nclust_range=[2, 7],
                                            s=classifier,
                                            c=clustering,
                                            nrand=100)
            metrics, nbest, _ = findbestclust.best_nclust(X_tr, iter_cv=10, strat_vect=y_tr)
            tmp_time = time.time() - start
            perf.append(tmp_time)
            print(f'Feat {f}, samples {s}: N cluster {nbest}, time: {tmp_time}')
        perftime.append(perf)

    perftime = np.array(perftime)
    fig, ax = plt.subplots()
    ax.plot(samples, perftime[:, 0], label='10 features', linestyle='--', color='black')
    ax.plot(samples, perftime[:, 1], label='100 features', color='black')
    ax.plot(samples, perftime[:, 2], label='500 features', linestyle='-.', color='black')
    ax.plot(samples, perftime[:, 3], label='1000 features', linestyle=':', color='black')
    ax.set_xlabel('Number of samples')
    ax.set_ylabel('Execution time (s)')
    ax.set_title("")
    ax.legend()
    plt.savefig('./performance_blobs.png', dpi=300)


if __name__ == "__main__":
    blobs_performance()
